/*  Autopackage GTK+ frontend
 *  Copyright (C) 2003  Hongli Lai <h.lai@chello.nl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>

#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <glade/glade.h>

#include "utils.h"
#include "package.h"
#include "prefs.h"
#include "prefix.h"

#define W(x) glade_xml_get_widget (xml, x)
static GladeXML *xml = NULL;
static gchar *filename = NULL;
static guint check_child_id = 0, progress_step_id = 0;
static gulong check_client_event_signal = 0;
static gchar *myAtomName = NULL;
static pid_t pid = 0;
static gboolean dontWrite = FALSE;

gchar *data_dir;
gchar *bin_dir;


static gboolean
check_client_event (GtkWidget *win, GdkEventClient *event)
{
	gchar *name = gdk_atom_name (event->message_type);

	if (name && strcmp (name, myAtomName) == 0
	  && event->data_format == 8 && strcmp (event->data.b, "DONE") == 0)
	{
		/* The frontend poked us; we can rest in peace now */

		gtk_widget_hide (W("main"));
		gtk_main_quit ();
		while (gtk_events_pending ()) gtk_main_iteration ();

		if (pid > 0)
		{
			int status;
			waitpid (pid, &status, 0);
			if (WIFEXITED (status))
				exit (WEXITSTATUS (status));
			else
				exit (1);
		} else
			exit (0);
	}
	return FALSE;
}


static void
write_pref ()
{
	if (dontWrite)
		return;
	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (W("silent"))))
		pref_set ("silent", "true");
	else
		pref_set ("silent", "false");
}


static void
initialize ()
{
	PackageInfo *info;
	GtkWidget *head, *author, *packager;

	if (!g_file_test (filename, G_FILE_TEST_EXISTS))
	{
		show_error (NULL, _("The given file was not found:\n%s"),
			filename);
		exit (1);
	}

	info = package_info_load (filename);
	if (!info)
	{
		show_error (NULL, _("The given file does not appear to "
			"be a valid Autopackage package:\n%s"),
			filename);
		exit (1);
	}

	head = W("head");
	author = W("author");
	packager = W("packager");

	if (info->summary && *info->summary)
	{
		gtk_label_set_text (GTK_LABEL (head), g_strdup_printf (
			_("%s (%s)\n"
			"Version %s\n\n"
			"%s"),
			info->displayName, info->shortName, info->softwareVersion,
			info->summary));
	} else
	{
		gtk_label_set_text (GTK_LABEL (head), g_strdup_printf (
			_("%s (%s)\n"
			"Version %s"),
			info->displayName, info->shortName, info->softwareVersion));
	}

	if (info->maintainer)
		gtk_label_set_text (GTK_LABEL (author), info->maintainer);
	else
	{
		gtk_widget_hide (W("label8"));
		gtk_widget_hide (author);
	}

	if (info->packager)
		gtk_label_set_text (GTK_LABEL (packager), info->packager);
	else
	{
		gtk_widget_hide (W("label9"));
		gtk_widget_hide (packager);
	}

	if (!info->maintainer && !info->packager)
		gtk_box_set_spacing (GTK_BOX (W("vbox2")), 0);

	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (W("silent")), pref_get_bool ("silent"));


	/* Reverse the button order if we're not in GNOME */
	if (!g_getenv ("GNOME_DESKTOP_SESSION_ID") || !*g_getenv ("GNOME_DESKTOP_SESSION_ID"))
	{
		reverse_container_children (GTK_CONTAINER (W("hbuttonbox2")));
		gtk_widget_grab_default (W("install"));
		gtk_widget_grab_focus (W("install"));
	}

	/* Check for client event from the frontend */
	if (!myAtomName)
		myAtomName = g_strdup_printf ("ApkgGtkFeDone-%d", getpid ());
	check_client_event_signal = g_signal_connect (W("main"), "client_event",
		G_CALLBACK (check_client_event), NULL);

	gtk_window_set_title (GTK_WINDOW (W("main")),
		g_strdup_printf (_("%s - Software installation"), info->displayName));
	gtk_widget_realize (W("main"));
	gtk_widget_show (W("main"));

	atexit (write_pref);
}


gboolean
help ()
{
	GdkColor *white;
	GladeXML *helpXML;
	gchar *version;
	#define H(x) glade_xml_get_widget (helpXML, x)

	helpXML = load_glade ("quickhelp.glade");
	gtk_window_set_transient_for (GTK_WINDOW (H("quickHelp")), GTK_WINDOW (W("main")));
	gtk_window_set_transient_for (GTK_WINDOW (H("aboutBox")), GTK_WINDOW (H("quickHelp")));
	gtk_window_set_type_hint (GTK_WINDOW (H("aboutBox")), GDK_WINDOW_TYPE_HINT_DIALOG);
	white = &gtk_widget_get_style (H("quickHelpEvent"))->white;
	gtk_widget_modify_bg (H("quickHelpEvent"), GTK_STATE_NORMAL, white);

	version = g_strdup_printf ("<span size=\"x-large\" weight=\"bold\">Autopackage %s</span>",
		VERSION);
	gtk_label_set_markup (GTK_LABEL (H("versionLabel")), version);
	g_free (version);

	gtk_widget_show (H("quickHelp"));
	g_object_unref (helpXML);
	return FALSE;
}


static gboolean
progress_step (gpointer data)
{
	GtkWidget *progress;

	progress = W("progress");
	gtk_progress_bar_pulse (GTK_PROGRESS_BAR (progress));
	return TRUE;
}


static gboolean
check_child (gpointer data)
{
	pid_t pid = GPOINTER_TO_INT (data);
	int status, result;

	result = waitpid (pid, &status, WNOHANG);
	if (result == -1)
	{
		/* Error? */
		g_source_remove (progress_step_id);
		g_source_remove (check_child_id);
		g_signal_handler_disconnect (W("main"), check_client_event_signal);

		kill (pid, SIGTERM);
		show_error (W("main"), _("An unknown error has occured."));
		waitpid (pid, &status, 0);
		exit (1);
	} else if (result == pid)
	{
		/* Child exited */
		g_source_remove (progress_step_id);
		g_source_remove (check_child_id);
		g_signal_handler_disconnect (W("main"), check_client_event_signal);

		if (WIFEXITED (status))
		{
			/* Normal exit */
			if (WEXITSTATUS (status) == 0)
			{
				GtkWidget *dialog = gtk_message_dialog_new (GTK_WINDOW (W("main")),
					GTK_DIALOG_MODAL,
					GTK_MESSAGE_INFO,
					GTK_BUTTONS_OK,
					_("Installation successful."));
				gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);
				gtk_dialog_run (GTK_DIALOG (dialog));

				exit (0);
			} else if (WEXITSTATUS (status) == 1)
				show_error (W("main"),
					_("Unable to extract the package's contents."));
			else if (WEXITSTATUS (status) == 2)
				show_error (W("main"),
					_("Integrity check failed. This package is damaged.\n"
					"- If you have obtained it via the Internet, please download it again.\n"
					"- Otherwise, or if this problem keeps occuring, contact this package's "
					"packager/vendor."));
			else
				show_error (W("main"),
					_("An unknown error has occured (exit status %d)."),
					WEXITSTATUS (status));

			exit (1);
		} else if (WIFSIGNALED (status))
		{
			/* Child crashed? */
			show_error (W("main"),
				_("The installation has failed because the package "
				"has encountered a system error.\n"
				"(Process %d caught signal %d (%s))"),
				pid, WTERMSIG (status),
				g_strsignal (WTERMSIG (status)));
			exit (1);
		}
	}
	return TRUE;
}


void
install_click ()
{
	/* Launch the package using bash.
	   Create a timeout for the progress bar pulse and
	   another one for checking the child's exit status
	   and whether the frontend has been launched yet. */

	gchar *frontend = get_frontend_filename ();
	write_pref ();
	dontWrite = TRUE;

	pid = fork ();
	switch (pid)
	{
	case 0:
		/* Child */
		setenv ("AUTOPACKAGE_GTKFE_ATOM", myAtomName, TRUE);
		setenv ("AUTOPACKAGE_GTKFE_XID",
			g_strdup_printf ("%d", (int) GDK_WINDOW_XWINDOW (W("main")->window)),
			TRUE);
		setenv ("AUTOPACKAGE_FRONTEND", frontend, TRUE);
		execl ("/bin/bash", "/bin/bash", filename, NULL);
		_exit (2);
	case -1:
		show_error (W("main"),
			_("There is too little free memory to launch this package."
			"Please close unnecessary programs and try again."));
		break;
	default:
		gtk_widget_set_sensitive (W("install"), FALSE);
		progress_step_id = gtk_timeout_add (10, progress_step, NULL);
		check_child_id = gtk_timeout_add (100, check_child, GINT_TO_POINTER (pid));
	}
}


gboolean
cancel ()
{
	if (check_child_id)
	{
		int status;

		/* The installation is still running. Kill the child. */
		kill (pid, SIGTERM);
		waitpid (pid, &status, 0);
	}

	gtk_main_quit ();
	return TRUE;
}


int
main (int argc, char *argv[])
{
	gtk_init (&argc, &argv);

	data_dir = g_strdup (DATADIR);
	bin_dir = g_strdup (BINDIR);

	if (argv[1])
	{
		if (strcmp (argv[1], "--version") == 0 || strcmp (argv[1], "-V") == 0)
		{
			g_print (_("Autopackage GTK+ frontend version %s\n"), VERSION);
			return 0;
		}

		xml = load_glade ("launcher.glade");
		filename = argv[1];
		initialize ();
		gtk_main ();
		return 0;
	} else
	{
		g_print ("Usage: autopackage-launcher-gtk <FILENAME>\n");
		return 1;
	}
}
